from enum import Enum

Wies = Enum('Wies', ['SEQUENCE_3', 'SEQUENCE_4', 'SEQUENCE_5', '4_SAME', '4_UNDER', '4_NAELL'])

points_wies = {Wies.SEQUENCE_3: 20, Wies.SEQUENCE_4: 50, Wies.SEQUENCE_5: 100, '4_NAELL': 150, '4_UNDER': 200}


# TODO: Implement Wiesen and the corresponding rules

def wies_allowed(wies, hand_cards):
    allowed = False
    if not len(wies) >= 3:
        return False
    if not set(wies) < set(hand_cards):
        return False
    return allowed



from enum import Enum
from operator import itemgetter

from pyschieber.suit import Suit
from pyschieber.trumpf import Trumpf
from pyschieber.rules.count_rules import counting_factor

# https://www.jassverzeichnis.ch/index.php/blog/95-jass-tipps-trumpfansagen-schieber
TrumpfType = Enum('TrumpfType',
                  ['UNDER_4', 'NELL_ASS_5', 'UNDER_NELL_ASS', 'UNDER_NELL_3_2_ASS', 'STICHE_5', 'NO_TRUMPF',
                   'HAVE_TO_DECIDE'])


def choose_trumpf(cards, geschoben):
    candidates = []
    for trumpf in filter(lambda x: x != Trumpf.OBE_ABE and x != Trumpf.UNDE_UFE and x != Trumpf.SCHIEBEN, Trumpf):
        trumpf_type = evalute_stich_trumpf(cards, trumpf.name)
        if not trumpf_type == TrumpfType.NO_TRUMPF:
            candidates.append((trumpf, trumpf_type))
    trumpf_type = evalute_unde_ufe(cards)
    if not trumpf_type == TrumpfType.NO_TRUMPF:
        candidates.append((Trumpf.UNDE_UFE, trumpf_type))
    trumpf_type = evalute_obe_abe(cards)
    if not trumpf_type == TrumpfType.NO_TRUMPF:
        candidates.append((Trumpf.OBE_ABE, trumpf_type))
    if len(candidates) >= 1:
        return choose_candidate(candidates)
    elif geschoben:
        return have_to_decide(cards)
    return Trumpf.SCHIEBEN, TrumpfType.NO_TRUMPF


def choose_candidate(candidates):
    max_index = 0
    max_factor = 0
    for i in range(0, len(candidates)):
        trumpf, _ = candidates[i]
        current_factor = counting_factor[trumpf]
        if max_factor < current_factor:
            max_factor = current_factor
            max_index = i
    return candidates[max_index]


def have_to_decide(cards):
    number_of_stiche = count_stiche_per_trumpf(cards)
    number_of_stiche.append((Trumpf.OBE_ABE, len(count_stiche(cards=cards, best_card=14, step=-1))))
    number_of_stiche.append((Trumpf.UNDE_UFE, len(count_stiche(cards=cards, best_card=6, step=1))))
    trumpf, count = max(number_of_stiche, key=itemgetter(1))
    return trumpf, TrumpfType.HAVE_TO_DECIDE


def evalute_stich_trumpf(cards, suit_name):
    trumpf_card_values = [card.value for card in cards if card.suit.name == suit_name]
    non_trumpf_card_values = [card.value for card in cards if card.suit.name != suit_name]
    nell = bool(9 in trumpf_card_values)
    under = bool(11 in trumpf_card_values)
    ass = bool(14 in trumpf_card_values)
    if under and len(trumpf_card_values) == 4:
        return TrumpfType.UNDER_4
    if ass and nell and len(trumpf_card_values) == 5:
        return TrumpfType.NELL_ASS_5
    if ass and nell and under:
        return TrumpfType.UNDER_NELL_ASS
    if nell and under and non_trumpf_card_values.count(14) >= 2:
        return TrumpfType.UNDER_NELL_3_2_ASS
    return TrumpfType.NO_TRUMPF


def evalute_unde_ufe(cards):
    stiche = count_stiche(cards=cards, best_card=6, step=1)
    return TrumpfType.STICHE_5 if len(stiche) > 5 else TrumpfType.NO_TRUMPF


def evalute_obe_abe(cards):
    stiche = count_stiche(cards=cards, best_card=14, step=-1)
    return TrumpfType.STICHE_5 if len(stiche) > 5 else TrumpfType.NO_TRUMPF


def count_stiche_per_trumpf(cards):
    stiche = []
    suits = split_cards_by_suit(cards)
    for suit, suit_cards in suits:
        nell = bool(9 in suit_cards)
        under = bool(11 in suit_cards)
        stich_counter = 0
        if under:
            stich_counter += 1
            if nell:
                stich_counter += 1
        stiche.append((Trumpf[suit.name], stich_counter))
    return stiche


def count_stiche(cards, best_card, step=1):
    init_best_card = best_card
    suits = split_cards_by_suit(cards)
    stiche = []
    for suit, suit_cards in suits:
        suit_card_copy = list(suit_cards)
        for _ in suit_cards:
            if best_card in suit_cards:
                stiche.append((suit, best_card))
                suit_card_copy.remove(best_card)
                best_card += step
            else:
                break
        if abs(init_best_card - best_card) >= 3:
            for card in suit_card_copy:
                stiche.append((suit, card))
        best_card = init_best_card
    return stiche


def split_cards_by_suit(cards):
    suits = []
    for suit in Suit:
        suit_cards = [card.value for card in cards if card.suit.name == suit.name]
        suits.append((suit, suit_cards))
    return suits



from collections import namedtuple

PlayedCard = namedtuple('PlayedCard', ['player', 'card'])
Stich = namedtuple('Stich', ['player', 'played_cards', 'trumpf'])


def played_cards_dict(played_card):
    return {
        'player_id': played_card.player.id,
        'card': str(played_card.card)
    }


def stich_dict(stich):
    return {
        'player_id': stich.player.id,
        'trumpf': stich.trumpf.name,
        'played_cards': [played_cards_dict(played_card) for played_card in stich.played_cards]
    }



from pyschieber.player.base_player import BasePlayer

from example.trumpf_decision import choose_trumpf


class GreedyPlayer(BasePlayer):
    bad_trumpf_reward = -100
    bad_card_reward = -10

    @classmethod
    def with_tournament(cls, name, tournament):
        player = cls(name=name)
        player.tournament = tournament
        player.reward = 0
        player.points_team_1 = 0
        player.points_team_2 = 0
        return player

    def choose_trumpf(self, geschoben):
        allowed = False
        while not allowed:
            trumpf, _ = choose_trumpf(cards=self.cards, geschoben=geschoben)
            allowed = yield trumpf
            if allowed:
                yield None
            else:
                self.reward += self.bad_trumpf_reward

    def choose_card(self, state=None):
        self.cards.sort()
        allowed = False
        last = -1
        while not allowed:
            card = self.cards[last]
            allowed = yield card
            if allowed:
                yield None
            else:
                last -= 1
                self.reward += self.bad_card_reward

    def stich_over(self, state=None):
        self.calculate_reward()

    def calculate_reward(self):
        teams = self.tournament.teams
        current_points_team_1 = teams[0].points
        current_points_team_2 = teams[1].points
        diff_1 = current_points_team_1 - self.points_team_1
        diff_2 = current_points_team_2 - self.points_team_2
        self.points_team_1, self.points_team_2 = current_points_team_1, current_points_team_2
        reward = diff_1 - diff_2
        self.reward += reward



class Team:
    def __init__(self, players=None):
        self.points = 0
        self.players = players

    def player_by_number(self, number):
        for player in self.players:
            if player.number == number:
                return player
        return None

    def won(self, point_limit):
        return self.points >= point_limit



import pytest

from pyschieber.rules.stich_rules import stich_rules, card_allowed, allowed_cards
from pyschieber.trumpf import Trumpf
from pyschieber.card import Card
from pyschieber.player.random_player import RandomPlayer
from pyschieber.stich import PlayedCard
from pyschieber.suit import Suit


@pytest.fixture(scope="module", autouse=True)
def players():
    return [RandomPlayer(), RandomPlayer(), RandomPlayer(), RandomPlayer()]


@pytest.fixture(scope="module", autouse=True)
def played_cards(players):
    return [PlayedCard(player=players[0], card=Card(Suit.BELL, 10)),
            PlayedCard(player=players[1], card=Card(Suit.ACORN, 6)),
            PlayedCard(player=players[2], card=Card(Suit.BELL, 13)),
            PlayedCard(player=players[3], card=Card(Suit.BELL, 9))]


@pytest.mark.parametrize("trumpf, index,", [
    (Trumpf.OBE_ABE, 2),
    (Trumpf.UNDE_UFE, 3),
    (Trumpf.BELL, 3),
    (Trumpf.ACORN, 1),
])
def test_stich(trumpf, index, players, played_cards):
    stich = stich_rules[trumpf](played_cards=played_cards)
    assert stich.player is players[index]


@pytest.mark.parametrize("first_card, chosen_card, hand_cards, trumpf, result", [
    (Card(Suit.ACORN, 12), Card(Suit.BELL, 12), [Card(Suit.ACORN, 11), Card(Suit.BELL, 12), Card(Suit.BELL, 11)],
     Trumpf.OBE_ABE, False),
    (Card(Suit.BELL, 12), Card(Suit.BELL, 12), [Card(Suit.BELL, 12), Card(Suit.BELL, 11)], Trumpf.BELL, True),
    (None, Card(Suit.BELL, 12), [Card(Suit.BELL, 12), Card(Suit.BELL, 11)], Trumpf.BELL, True),
    (Card(Suit.BELL, 12), Card(Suit.BELL, 12), [Card(Suit.BELL, 12), Card(Suit.BELL, 11)], Trumpf.OBE_ABE, True),
    (Card(Suit.ACORN, 12), Card(Suit.BELL, 12), [Card(Suit.BELL, 12), Card(Suit.BELL, 11)], Trumpf.OBE_ABE, True),
    (Card(Suit.ACORN, 11), Card(Suit.BELL, 12), [Card(Suit.ACORN, 12), Card(Suit.BELL, 11), Card(Suit.BELL, 12)],
     Trumpf.ACORN, True),
    (Card(Suit.ACORN, 11), Card(Suit.ACORN, 11), [Card(Suit.BELL, 12), Card(Suit.ACORN, 11), Card(Suit.ACORN, 12)],
     Trumpf.UNDE_UFE, True),
    (Card(Suit.ACORN, 11), Card(Suit.ACORN, 11), [Card(Suit.BELL, 12), Card(Suit.ACORN, 12)],
     Trumpf.UNDE_UFE, False),
])
def test_card_allowed(first_card, chosen_card, hand_cards, trumpf, result):
    assert card_allowed(first_card, chosen_card, hand_cards, trumpf) == result


@pytest.mark.parametrize("hand_cards, table_cards, trumpf, result", [
    ([Card(Suit.BELL, 12)], [Card(Suit.BELL, 11)], Trumpf.BELL, [Card(Suit.BELL, 12)]),
    ([Card(Suit.BELL, 12), Card(Suit.ROSE, 12)], [Card(Suit.BELL, 11)], Trumpf.ACORN, [Card(Suit.BELL, 12)]),
    ([Card(Suit.BELL, 12)], [Card(Suit.ROSE, 11)], Trumpf.ACORN, [Card(Suit.BELL, 12)]),
])
def test_allowed_cards(hand_cards, table_cards, trumpf, result):
    assert allowed_cards(hand_cards=hand_cards, table_cards=table_cards, trumpf=trumpf) == result



import random

from pyschieber.card import from_string_to_card
from pyschieber.player.base_player import BasePlayer
from pyschieber.trumpf import Trumpf


class RandomPlayer(BasePlayer):
    def choose_trumpf(self, geschoben):
        return move(choices=list(Trumpf))

    def choose_card(self, state=None):
        cards = self.allowed_cards(state=state)
        return move(choices=cards)

def move(choices):
    allowed = False
    while not allowed:
        choice = random.choice(choices)
        allowed = yield choice
        if allowed:
            yield None



from example.greedy_player import GreedyPlayer
from pyschieber.player.random_player import RandomPlayer
from pyschieber.tournament import Tournament


def test_example_player():
    tournament = Tournament(point_limit=1500)
    example_player = GreedyPlayer.with_tournament(name='ExamplePlayer', tournament=tournament)
    tournament.register_player(example_player)
    [tournament.register_player(RandomPlayer(name=str(i))) for i in range(2, 5)]
    tournament.play()
    assert example_player.reward != 0




from pyschieber.player.base_player import BasePlayer
from itertools import count


def test_base_player_counter():
    BasePlayer.class_counter = count(0)
    base_player = [BasePlayer(name=i) for i in range(4)]

    for i, player in enumerate(base_player):
        assert player.name == i



import pytest

from example.trumpf_decision import choose_trumpf, TrumpfType
from pyschieber.card import Card
from pyschieber.suit import Suit
from pyschieber.trumpf import Trumpf


@pytest.mark.parametrize("cards, trumpf, trumpf_type", [
    ([Card(Suit.BELL, 13), Card(Suit.BELL, 12), Card(Suit.BELL, 11), Card(Suit.BELL, 8), Card(Suit.ACORN, 12),
      Card(Suit.ACORN, 11), Card(Suit.SHIELD, 12), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.BELL,
     TrumpfType.UNDER_4),
    ([Card(Suit.BELL, 14), Card(Suit.BELL, 12), Card(Suit.BELL, 9), Card(Suit.BELL, 8), Card(Suit.BELL, 6),
      Card(Suit.ACORN, 11), Card(Suit.SHIELD, 12), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.BELL,
     TrumpfType.NELL_ASS_5),
    ([Card(Suit.BELL, 14), Card(Suit.BELL, 11), Card(Suit.BELL, 9), Card(Suit.ACORN, 12), Card(Suit.ACORN, 12),
      Card(Suit.ACORN, 7), Card(Suit.SHIELD, 12), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.BELL,
     TrumpfType.UNDER_NELL_ASS),
    ([Card(Suit.BELL, 11), Card(Suit.BELL, 9), Card(Suit.BELL, 6), Card(Suit.ACORN, 14), Card(Suit.ACORN, 12),
      Card(Suit.ACORN, 7), Card(Suit.SHIELD, 14), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.BELL,
     TrumpfType.UNDER_NELL_3_2_ASS),
    ([Card(Suit.BELL, 14), Card(Suit.BELL, 13), Card(Suit.BELL, 12), Card(Suit.BELL, 10), Card(Suit.BELL, 6),
      Card(Suit.ACORN, 7), Card(Suit.SHIELD, 14), Card(Suit.ROSE, 13), Card(Suit.ROSE, 12)], Trumpf.OBE_ABE,
     TrumpfType.STICHE_5),
    ([Card(Suit.BELL, 6), Card(Suit.BELL, 7), Card(Suit.BELL, 8), Card(Suit.BELL, 10), Card(Suit.ACORN, 6),
      Card(Suit.ACORN, 7), Card(Suit.SHIELD, 14), Card(Suit.ROSE, 13), Card(Suit.ROSE, 12)], Trumpf.UNDE_UFE,
     TrumpfType.STICHE_5),
])
def test_choose_trumpf_no_schieben(cards, trumpf, trumpf_type):
    evaluate_trumpf, evaluate_trumpf_type = choose_trumpf(cards=cards, geschoben=False)
    assert (evaluate_trumpf, evaluate_trumpf_type) == (trumpf, trumpf_type)


@pytest.mark.parametrize("cards, trumpf, trumpf_type", [
    ([Card(Suit.BELL, 10), Card(Suit.BELL, 12), Card(Suit.BELL, 6), Card(Suit.BELL, 8), Card(Suit.ACORN, 12),
      Card(Suit.ACORN, 11), Card(Suit.SHIELD, 12), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.ACORN,
     TrumpfType.HAVE_TO_DECIDE),
    ([Card(Suit.BELL, 10), Card(Suit.BELL, 12), Card(Suit.BELL, 6), Card(Suit.BELL, 7), Card(Suit.ACORN, 12),
      Card(Suit.ACORN, 11), Card(Suit.SHIELD, 12), Card(Suit.SHIELD, 7), Card(Suit.ROSE, 9)], Trumpf.UNDE_UFE,
     TrumpfType.HAVE_TO_DECIDE),
])
def test_choose_trumpf_schieben(cards, trumpf, trumpf_type):
    evaluate_trumpf, evaluate_trumpf_type = choose_trumpf(cards=cards, geschoben=True)
    assert (evaluate_trumpf, evaluate_trumpf_type) == (trumpf, trumpf_type)



from random import shuffle

from pyschieber.deck import Deck


class Dealer:
    def __init__(self, players):
        self.players = players
        self.deck = Deck()

    def shuffle_cards(self):
        shuffle(self.deck.cards)

    def deal_cards(self):
        for i, card in enumerate(self.deck.cards):
            self.players[i % 4].set_card(card=card)



from functools import partial

from pyschieber.stich import Stich
from pyschieber.trumpf import Trumpf

UNDER = 11
NAELL = 9


def stich_obe_unde(played_cards, operation, trumpf):
    suit = played_cards[0].card.suit
    (_, index) = operation(
        [(played_card.card.value, i) for i, played_card in enumerate(played_cards) if played_card.card.suit == suit])
    return Stich(player=played_cards[index].player, played_cards=played_cards, trumpf=trumpf)


def stich_trumpf(played_cards, trumpf):
    trumpfs = [(played_card.card.value, i) for i, played_card in enumerate(played_cards) if
               played_card.card.suit.name == trumpf.name]
    if trumpfs:
        values = [trumpf[0] for trumpf in trumpfs]
        if UNDER in values:  # Under
            index = trumpfs[values.index(UNDER)][1]
            return Stich(player=played_cards[index].player, played_cards=played_cards, trumpf=trumpf)
        if NAELL in values:
            index = trumpfs[values.index(NAELL)][1]
            return Stich(player=played_cards[index].player, played_cards=played_cards, trumpf=trumpf)
        index = max(trumpfs)[1]
        return Stich(player=played_cards[index].player, played_cards=played_cards, trumpf=trumpf)
    else:
        return stich_obe_unde(played_cards=played_cards, operation=max, trumpf=trumpf)


stich_rules = {
    Trumpf.OBE_ABE: partial(stich_obe_unde, operation=max, trumpf=Trumpf.OBE_ABE),
    Trumpf.UNDE_UFE: partial(stich_obe_unde, operation=min, trumpf=Trumpf.UNDE_UFE),
}

for trumpf in filter(lambda x: x != Trumpf.OBE_ABE and x != Trumpf.UNDE_UFE and x != Trumpf.SCHIEBEN, Trumpf):
    stich_rules[trumpf] = partial(stich_trumpf, trumpf=trumpf)


def card_allowed(first_card, chosen_card, hand_cards, trumpf):
    chosen_suit = chosen_card.suit
    if chosen_card not in hand_cards:
        return False
    if first_card is None:
        return True
    first_suit = first_card.suit
    if first_suit == chosen_suit or chosen_suit.name == trumpf.name:
        return True
    else:
        if trumpf not in [Trumpf.OBE_ABE, Trumpf.UNDE_UFE]:
            hand_suits = [hand_card.suit for hand_card in hand_cards if
                          hand_card.suit.name != trumpf.name and hand_card.value != UNDER]
        else:
            hand_suits = [hand_card.suit for hand_card in hand_cards]
        if first_suit in hand_suits:
            return False
        else:
            return True


def allowed_cards(hand_cards, table_cards, trumpf):
    cards = []
    if len(table_cards) > 0:
        for card in hand_cards:
            if card_allowed(table_cards[0], card, hand_cards, trumpf):
                cards.append(card)
    else:
        cards += hand_cards
    return cards



import pytest

from pyschieber.deck import Deck
from pyschieber.trumpf import Trumpf
from pyschieber.rules.count_rules import count_stich


@pytest.mark.parametrize("trumpf", list(Trumpf)[:6])
def test_count_rules_with_last(trumpf):
    deck = Deck()
    points = count_stich(cards=deck.cards, trumpf=trumpf, last=True)
    assert points == 157


@pytest.mark.parametrize("trumpf", list(Trumpf)[:6])
def test_count_rules(trumpf):
    deck = Deck()
    points = count_stich(cards=deck.cards, trumpf=trumpf, last=False)
    assert points == 152



from pyschieber.suit import Suit
from pyschieber.card import Card


class Deck:
    def __init__(self):
        self.cards = []
        for suit in Suit:
            self.cards += [Card(suit=suit, value=i) for i in range(6, 15)]

    def __str__(self):
        return str([str(card) for card in self.cards])



from pyschieber.player.base_player import BasePlayer


class CliPlayer(BasePlayer):
    def choose_trumpf(self, geschoben=False):
        self._print_cards()
        print('\nTrumpf:')
        for i, trumpf in enumerate(self.trumpf_list):
            print('{0} : {1}'.format(i, trumpf))
        print('\nGeschoben: {0}\n'.format(geschoben))
        return move_allowed(move_function=self._choose_trumpf_input, message="Schieben not allowed!\n")

    def choose_card(self, state=None):
        self._print_cards()
        return move_allowed(move_function=self._choose_card_input, message="Card not allowed!\n")

    def _choose_trumpf_input(self):
        while True:
            try:
                trumpf_index = int(
                    input("Please chose the trumpf by the number from 0 to {0}: \n".format(len(self.trumpf_list) - 1)))
                if trumpf_index in range(0, len(self.trumpf_list)):
                    return self.trumpf_list[trumpf_index]
                else:
                    print("Sorry, no valid trumpf number.\n")
                    continue
            except ValueError:
                print("Sorry, I didn't understand that.\n")
                continue

    def _choose_card_input(self):
        while True:
            try:
                card_index = int(
                    input("Please chose the card by the number from 0 to {0}: \n".format(len(self.cards) - 1)))
                if card_index in range(0, len(self.cards)):
                    print()
                    return self.cards[card_index]
                else:
                    print("Sorry, no valid trumpf number\n")
                    continue
            except ValueError:
                print("Sorry, I didn't understand that.\n")
                continue

    def _print_cards(self):
        print('Hand cards: \n')
        for i, card in enumerate(self.cards):
            print('{0} : {1}'.format(i, card))
        print('')


def move_allowed(move_function, message):
    allowed = False
    while not allowed:
        move = move_function()
        allowed = yield move
        if allowed:
            yield None
        else:
            print(message)



import inspect

from pyschieber.card import from_string_to_card
from pyschieber.trumpf import Trumpf
from pyschieber.rules.stich_rules import allowed_cards


class BasePlayer:
    def __init__(self, name='unknown'):
        self.name = name
        self.cards = []
        self.trumpf_list = list(Trumpf)
        self.id = None

    def get_dict(self):
        return dict(name=self.name, type=type(self).__name__)

    def set_card(self, card):
        self.cards.append(card)

    def choose_trumpf(self, geschoben):
        raise NotImplementedError(str(inspect.stack()[1][3]))

    def choose_card(self, state=None):
        raise NotImplementedError(str(inspect.stack()[1][3]))

    def stich_over(self, state=None):
        pass

    def allowed_cards(self, state):
        table_cards = [from_string_to_card(entry['card']) for entry in state['table']]
        trumpf = Trumpf[state['trumpf']]
        return allowed_cards(hand_cards=self.cards, table_cards=table_cards, trumpf=trumpf)

    def __str__(self):
        return '<Player:{}>'.format(self.name)



import pytest

from pyschieber.trumpf import Trumpf
from pyschieber.rules.trumpf_rules import trumpf_allowed


@pytest.mark.parametrize("trumpf, geschoben, result", [
    (Trumpf.OBE_ABE, True, True),
    (Trumpf.OBE_ABE, False, True),
    (Trumpf.SCHIEBEN, True, False),
    (Trumpf.SCHIEBEN, False, True),
])
def test_trumpf(trumpf, geschoben, result):
    assert trumpf_allowed(chosen_trumpf=trumpf, geschoben=geschoben) == result



from pyschieber.player.random_player import RandomPlayer
from pyschieber.tournament import Tournament
from example.greedy_player import GreedyPlayer


def start_tournament(points):
    tournament = Tournament(point_limit=points)

    players = [RandomPlayer(name='Tick'), RandomPlayer(name='Trick'), RandomPlayer(name='Track'),
               GreedyPlayer.with_tournament(name='Dagobert', tournament=tournament)]

    [tournament.register_player(player) for player in players]

    tournament.play()


if __name__ == "__main__":
    start_tournament(points=1500)



import pytest
from timeit import default_timer as timer
from math import sqrt, floor

from pyschieber.player.random_player import RandomPlayer
from pyschieber.tournament import Tournament


@pytest.mark.statistical
def test_is_random():
    point_limit = 1000
    number_of_tournaments = 1000
    mean = number_of_tournaments * 0.5  # assume that a RandomPlayer has a 50% chance to win
    variance = mean * (1 - 0.5)
    standard_deviation = int(floor(sqrt(variance)))

    random_players = [RandomPlayer(name=i) for i in range(4)]
    tournament = Tournament(point_limit=point_limit)
    [tournament.register_player(player=player) for player in random_players]

    team_1_won = 0
    team_2_won = 0

    start = timer()

    for _ in range(number_of_tournaments):
        tournament.play()
        if tournament.teams[0].won(point_limit=point_limit):
            team_1_won += 1
        else:
            team_2_won += 1

    end = timer()
    print("\nTo run {0} tournaments it took {1:.2f} seconds.".format(number_of_tournaments, end - start))

    difference = abs(team_1_won - team_2_won)
    print("Difference: ", difference)
    print("Team 1: ", team_1_won)
    print("Team 2: ", team_2_won)
    assert difference in range(0, 2 * standard_deviation)



from pyschieber.trumpf import Trumpf

counting_factor = {Trumpf.ROSE: 1, Trumpf.ACORN: 1, Trumpf.BELL: 2,  Trumpf.SHIELD: 2, Trumpf.OBE_ABE: 3,
                   Trumpf.UNDE_UFE: 3}

points_obe_abe = {6: 0, 7: 0, 8: 8, 9: 0, 10: 10, 11: 2, 12: 3, 13: 4, 14: 11}
points_unde_ufe = {6: 11, 7: 0, 8: 8, 9: 0, 10: 10, 11: 2, 12: 3, 13: 4, 14: 0}
points_trumpf_color = {6: 0, 7: 0, 8: 0, 9: 14, 10: 10, 11: 20, 12: 3, 13: 4, 14: 11}
points_non_trumpf_color = {6: 0, 7: 0, 8: 0, 9: 0, 10: 10, 11: 2, 12: 3, 13: 4, 14: 11}

card_points = {Trumpf.OBE_ABE: points_obe_abe, Trumpf.UNDE_UFE: points_unde_ufe}

for trumpf in filter(lambda x: x != Trumpf.OBE_ABE and x != Trumpf.UNDE_UFE and x != Trumpf.SCHIEBEN, Trumpf):
    card_points[trumpf] = points_trumpf_color


def count_stich(cards, trumpf, last=False):
    points = 0 if not last else 5
    for card in cards:
        if trumpf == Trumpf.OBE_ABE or trumpf == Trumpf.UNDE_UFE or card.suit.name == trumpf.name:
            points += card_points[trumpf][card.value]
        else:
            points += points_non_trumpf_color[card.value]
    return points



import re

from pyschieber.suit import Suit


class Card:
    names = {6: '6', 7: '7', 8: '8', 9: '9', 10: 'Banner', 11: 'Under', 12: 'Ober', 13: 'Koennig', 14: 'Ass'}
    format_string = '<{0}:{1}>'

    def __init__(self, suit, value):
        self.suit = suit
        self.value = value

    def __lt__(self, other):
        return self.value < other.value

    def __eq__(self, other):
        return self.suit == other.suit and self.value == other.value

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        name = str(self.value)
        if self.value > 9:
            name = Card.names[self.value]
        return self.format_string.format(self.suit.name, name)

    def __repr__(self):
        return str(self)


def from_string_to_card(card_string):
    regex = re.sub(r'{(.+?)}', r'(?P<_\1>.+)', Card.format_string)
    values = list(re.search(regex, card_string).groups())
    suit = Suit[values[0]]
    card_value = ''
    for key, value in Card.names.items():
        if value == values[1]:
            card_value = key
            break
    return Card(suit=suit, value=card_value)
